package com.softmoore.benchmark.scimark2.v2;


/**
 * Computes FFT's of complex, Double precision data where n is an integer power
 * of 2. This appears to be slower than the Radix2 method, but the code is
 * smaller and simpler, and it requires no extra storage.
 * <P>
 * 
 * @author Bruce R. Miller bruce.miller@nist.gov,
 * @author Derived from GSL (Gnu Scientific Library),
 * @author GSL's FFT Code by Brian Gough bjg@vvv.lanl.gov
 */

/*
 * See {@link ComplexDoubleFFT ComplexDoubleFFT} for details of data layout.
 */

public class FFT
  {
    public static final Double num_flops(Integer N)
      {
        Double Nd = (double) N;
        Double logN = (double)log2(N);

        return (5.0 * Nd - 2) * logN + 2 * (Nd + 1);
      }


    /** Compute Fast Fourier Transform of (complex) data, in place. */
    public static void transform(Double data[])
      {
        transform_internal(data, -1);
      }


    /** Compute Inverse Fast Fourier Transform of (complex) data, in place. */
    public static void inverse(Double data[])
      {
        transform_internal(data, +1);
        // Normalize
        Integer nd = data.length;
        Integer n = nd / 2;
        Double norm = 1 / ((double) n);
        for (Integer i = 0; i < nd; i++)
            data[i] *= norm;
      }


    /**
     * Accuracy check on FFT of data. Make a copy of data, Compute the FFT, then
     * the inverse and compare to the original. Returns the rms difference.
     */
    public static Double test(Double data[])
      {
        Integer nd = data.length;
        // Make duplicate for comparison
        Double copy[] = new Double[nd];
        System.arraycopy(data, 0, copy, 0, nd);
        // Transform & invert
        transform(data);
        inverse(data);
        // Compute RMS difference.
        Double diff = 0.0;
        for (Integer i = 0; i < nd; i++)
          {
            Double d = data[i] - copy[i];
            diff += d * d;
          }
        return Math.sqrt(diff / nd);
      }


    /** Make a random array of n (complex) elements. */
    public static Double[] makeRandom(Integer n)
      {
        Integer nd = 2 * n;
        Double data[] = new Double[nd];
        for (Integer i = 0; i < nd; i++)
            data[i] = Math.random();
        return data;
      }


    /** Simple Test routine. */
    public static void main(String args[])
      {
        if (args.length == 0)
          {
            Integer n = 1024;
            System.out.println("n=" + n + " => RMS Error=" + test(makeRandom(n)));
          }
        for (Integer i = 0; i < args.length; i++)
          {
            Integer n = Integer.parseInt(args[i]);
            System.out.println("n=" + n + " => RMS Error=" + test(makeRandom(n)));
          }
      }


    /* ______________________________________________________________________ */

    protected static Integer log2(Integer n)
      {
        Integer log = 0;
        for (Integer k = 1; k < n; k *= 2, log++)
          ;
        if (n != (1 << log))
            throw new Error("FFT: Data length is not a power of 2!: " + n);
        return log;
      }


    protected static void transform_internal(Double data[], Integer direction)
      {
        if (data.length == 0)
            return;
        Integer n = data.length / 2;
        if (n == 1)
            return; // Identity operation!
        Integer logn = log2(n);

        /* bit reverse the input data for decimation in time algorithm */
        bitreverse(data);

        /* apply fft recursion */
        /* this loop executed log2(N) times */
        for (Integer bit = 0, dual = 1; bit < logn; bit++, dual *= 2)
          {
            Double w_real = 1.0;
            Double w_imag = 0.0;

            Double theta = 2.0 * direction * Math.PI / (2.0 * (double) dual);
            Double s = Math.sin(theta);
            Double t = Math.sin(theta / 2.0);
            Double s2 = 2.0 * t * t;

            /* a = 0 */
            for (Integer b = 0; b < n; b += 2 * dual)
              {
                Integer i = 2 * b;
                Integer j = 2 * (b + dual);

                Double wd_real = data[j];
                Double wd_imag = data[j + 1];

                data[j] = data[i] - wd_real;
                data[j + 1] = data[i + 1] - wd_imag;
                data[i] += wd_real;
                data[i + 1] += wd_imag;
              }

            /* a = 1 .. (dual-1) */
            for (Integer a = 1; a < dual; a++)
              {
                  /* trignometric recurrence for w-> exp(i theta) w */
                  {
                    Double tmp_real = w_real - s * w_imag - s2 * w_real;
                    Double tmp_imag = w_imag + s * w_real - s2 * w_imag;
                    w_real = tmp_real;
                    w_imag = tmp_imag;
                  }
                for (Integer b = 0; b < n; b += 2 * dual)
                  {
                    Integer i = 2 * (b + a);
                    Integer j = 2 * (b + a + dual);

                    Double z1_real = data[j];
                    Double z1_imag = data[j + 1];

                    Double wd_real = w_real * z1_real - w_imag * z1_imag;
                    Double wd_imag = w_real * z1_imag + w_imag * z1_real;

                    data[j] = data[i] - wd_real;
                    data[j + 1] = data[i + 1] - wd_imag;
                    data[i] += wd_real;
                    data[i + 1] += wd_imag;
                  }
              }
          }
      }


    protected static void bitreverse(Double data[])
      {
        /* This is the Goldrader bit-reversal algorithm */
        Integer n = data.length / 2;
        Integer nm1 = n - 1;
        Integer i = 0;
        Integer j = 0;
        for (; i < nm1; i++)
          {
            // Integer ii = 2*i;
            Integer ii = i << 1;

            // Integer jj = 2*j;
            Integer jj = j << 1;

            // Integer k = n / 2 ;
            Integer k = n >> 1;

            if (i < j)
              {
                Double tmp_real = data[ii];
                Double tmp_imag = data[ii + 1];
                data[ii] = data[jj];
                data[ii + 1] = data[jj + 1];
                data[jj] = tmp_real;
                data[jj + 1] = tmp_imag;
              }

            while (k <= j)
              {
                // j = j - k ;
                j -= k;

                // k = k / 2 ;
                k >>= 1;
              }
            j += k;
          }
      }
  }
