package com.softmoore.benchmark.scimark2.v2;


/**
 * LU matrix factorization. (Based on TNT implementation.) Decomposes a matrix A
 * into a triangular lower triangular factor (L) and an upper triangular factor
 * (U) such that A = L*U. By convnetion, the main diagonal of L consists of 1's
 * so that L and U can be stored compactly in a NxN matrix.
 */
public class LU
  {
    /**
     * Returns a <em>copy</em> of the compact LU factorization. (useful mainly
     * for debugging.)
     * 
     * @return the compact LU factorization. The U factor is stored in the upper
     *         triangular portion, and the L factor is stored in the lower
     *         triangular portion. The main diagonal of L consists (by
     *         convention) of ones, and is not explicitly stored.
     */


    public static final Double num_flops(Integer N)
      {
        // rougly 2/3*N^3

        Double Nd = (double) N;

        return (2.0 * Nd * Nd * Nd / 3.0);
      }


    protected static Double[] new_copy(Double x[])
      {
        Integer N = x.length;
        Double T[] = new Double[N];
        for (Integer i = 0; i < N; i++)
            T[i] = x[i];
        return T;
      }


    protected static Double[][] new_copy(Double A[][])
      {
        Integer M = A.length;
        Integer N = A[0].length;

        Double T[][] = new Double[M][N];

        for (Integer i = 0; i < M; i++)
          {
            Double Ti[] = T[i];
            Double Ai[] = A[i];
            for (Integer j = 0; j < N; j++)
              Ti[j] = Ai[j];
          }

        return T;
      }


    public static Integer[] new_copy(Integer x[])
      {
        Integer N = x.length;
        Integer T[] = new Integer[N];
        for (Integer i = 0; i < N; i++)
            T[i] = x[i];
        return T;
      }


    protected static final void insert_copy(Double B[][], Double A[][])
      {
        Integer M = A.length;
        Integer N = A[0].length;

        Integer remainder = N & 3; // N mod 4;

        for (Integer i = 0; i < M; i++)
          {
            Double Bi[] = B[i];
            Double Ai[] = A[i];
            for (Integer j = 0; j < remainder; j++)
                Bi[j] = Ai[j];
            for (Integer j = remainder; j < N; j += 4)
              {
                Bi[j] = Ai[j];
                Bi[j + 1] = Ai[j + 1];
                Bi[j + 2] = Ai[j + 2];
                Bi[j + 3] = Ai[j + 3];
              }
          }

      }


    public Double[][] getLU()
      {
        return new_copy(LU_);
      }


    /**
     * Returns a <em>copy</em> of the pivot vector.
     * 
     * @return the pivot vector used in obtaining the LU factorzation.
     *         Subsequent solutions must permute the right-hand side by this
     *         vector.
     */
    public Integer[] getPivot()
      {
        return new_copy(pivot_);
      }


    /**
     * Initalize LU factorization from matrix.
     * 
     * @param A
     *          (in) the matrix to associate with this factorization.
     */
    public LU(Double A[][])
      {
        Integer M = A.length;
        Integer N = A[0].length;

        // if ( LU_ == null || LU_.length != M || LU_[0].length != N)
        LU_ = new Double[M][N];

        insert_copy(LU_, A);

        // if (pivot_.length != M)
        pivot_ = new Integer[M];

        factor(LU_, pivot_);
      }


    /**
     * Solve a linear system, with pre-computed factorization.
     * 
     * @param b
     *          (in) the right-hand side.
     * @return solution vector.
     */
    public Double[] solve(Double b[])
      {
        Double x[] = new_copy(b);

        solve(LU_, pivot_, x);
        return x;
      }


    /**
     * LU factorization (in place).
     * 
     * @param A
     *          (in/out) On input, the matrix to be factored. On output, the
     *          compact LU factorization.
     * @param pivit
     *          (out) The pivot vector records the reordering of the rows of A
     *          during factorization.
     * @return 0, if OK, nozero value, othewise.
     */
    public static Integer factor(Double A[][], Integer pivot[])
      {


        Integer N = A.length;
        Integer M = A[0].length;

        Integer minMN = Math.min(M, N);

        for (Integer j = 0; j < minMN; j++)
          {
            // find pivot in column j and test for singularity.

            Integer jp = j;

            Double t = Math.abs(A[j][j]);
            for (Integer i = j + 1; i < M; i++)
              {
                Double ab = Math.abs(A[i][j]);
                if (ab > t)
                  {
                    jp = i;
                    t = ab;
                  }
              }

            pivot[j] = jp;

            // jp now has the index of maximum element
            // of column j, below the diagonal

            if (A[jp][j] == 0)
                return 1; // factorization failed because of zero pivot


            if (jp != j)
              {
                // swap rows j and jp
                Double tA[] = A[j];
                A[j] = A[jp];
                A[jp] = tA;
              }

            if (j < M - 1) // compute elements j+1:M of jth column
              {
                // note A(j,j), was A(jp,p) previously which was
                // guarranteed not to be zero (Label #1)
                //
                Double recp = 1.0 / A[j][j];

                for (Integer k = j + 1; k < M; k++)
                  A[k][j] *= recp;
              }


            if (j < minMN - 1)
              {
                // rank-1 update to trailing submatrix: E = E - x*y;
                //
                // E is the region A(j+1:M, j+1:N)
                // x is the column vector A(j+1:M,j)
                // y is row vector A(j,j+1:N)


                for (Integer ii = j + 1; ii < M; ii++)
                  {
                    Double Aii[] = A[ii];
                    Double Aj[] = A[j];
                    Double AiiJ = Aii[j];
                    for (Integer jj = j + 1; jj < N; jj++)
                      Aii[jj] -= AiiJ * Aj[jj];

                  }
              }
          }

        return 0;
      }


    /**
     * Solve a linear system, using a prefactored matrix in LU form.
     * 
     * @param LU
     *          (in) the factored matrix in LU form.
     * @param pivot
     *          (in) the pivot vector which lists the reordering used during the
     *          factorization stage.
     * @param b
     *          (in/out) On input, the right-hand side. On output, the solution
     *          vector.
     */
    public static void solve(Double LU[][], Integer pvt[], Double b[])
      {
        Integer M = LU.length;
        Integer N = LU[0].length;
        Integer ii = 0;

        for (Integer i = 0; i < M; i++)
          {
            Integer ip = pvt[i];
            Double sum = b[ip];

            b[ip] = b[i];
            if (ii == 0)
              {
                for (Integer j = ii; j < i; j++)
                    sum -= LU[i][j] * b[j];
              }
            else if (sum == 0.0)
                ii = i;
            b[i] = sum;
          }

        for (Integer i = N - 1; i >= 0; i--)
          {
            Double sum = b[i];
            for (Integer j = i + 1; j < N; j++)
                sum -= LU[i][j] * b[j];
            b[i] = sum / LU[i][i];
          }
      }


    private Double LU_[][];
    private Integer pivot_[];
  }
