package com.softmoore.benchmark.scimark2.v2;


public class kernel
  {
    // each measurement returns approx Mflops

    public static Double measureFFT(Integer N, Double mintime, Random R)
      {
        // initialize FFT data as complex (N real/img pairs)

        Double x[] = RandomVector(2 * N, R);
        Double oldx[] = NewVectorCopy(x);
        long cycles = 1;
        Stopwatch Q = new Stopwatch();

        while (true)
          {
            Q.start();
            for (Integer i = 0; i < cycles; i++)
              {
                FFT.transform(x); // forward transform
                FFT.inverse(x); // backward transform
              }
            Q.stop();
            if (Q.read() >= mintime)
                break;

            cycles *= 2;
          }
        // approx Mflops

        final Double EPS = 1.0e-10;
        if (FFT.test(x) / N > EPS)
            return 0.0;

        return FFT.num_flops(N) * cycles / Q.read() * 1.0e-6;
      }


    public static Double measureSOR(Integer N, Double min_time, Random R)
      {
        Double G[][] = RandomMatrix(N, N, R);

        Stopwatch Q = new Stopwatch();
        Integer cycles = 1;
        while (true)
          {
            Q.start();
            SOR.execute(1.25, G, cycles);
            Q.stop();
            if (Q.read() >= min_time)
              break;

            cycles *= 2;
          }
        // approx Mflops
        return SOR.num_flops(N, N, cycles) / Q.read() * 1.0e-6;
      }


    public static Double measureMonteCarlo(Double min_time, Random R)
      {
        Stopwatch Q = new Stopwatch();

        Integer cycles = 1;
        while (true)
          {
            Q.start();
            MonteCarlo.integrate(cycles);
            Q.stop();
            if (Q.read() >= min_time)
              break;

            cycles *= 2;
          }
        // approx Mflops
        return MonteCarlo.num_flops(cycles) / Q.read() * 1.0e-6;
      }


    public static Double measureSparseMatmult(Integer N, Integer nz, Double min_time, Random R)
      {
        // initialize vector multipliers and storage for result
        // y = A*y;

        Double x[] = RandomVector(N, R);
        Double y[] = new Double[N];

        // initialize square sparse matrix
        //
        // for this test, we create a sparse matrix wit M/nz nonzeros
        // per row, with spaced-out evenly between the begining of the
        // row to the main diagonal. Thus, the resulting pattern looks
        // like
        // +-----------------+
        // +* +
        // +*** +
        // +* * * +
        // +** * * +
        // +** * * +
        // +* * * * +
        // +* * * * +
        // +* * * * +
        // +-----------------+
        //
        // (as best reproducible with integer artihmetic)
        // Note that the first nr rows will have elements past
        // the diagonal.

        Integer nr = nz / N; // average number of nonzeros per row
        Integer anz = nr * N; // _actual_ number of nonzeros


        Double val[] = RandomVector(anz, R);
        Integer col[] = new Integer[anz];
        Integer row[] = new Integer[N + 1];

        row[0] = 0;
        for (Integer r = 0; r < N; r++)
          {
            // initialize elements for row r

            Integer rowr = row[r];
            row[r + 1] = rowr + nr;
            Integer step = r / nr;
            if (step < 1)
                step = 1; // take at least unit steps


            for (Integer i = 0; i < nr; i++)
                col[rowr + i] = i * step;

          }

        Stopwatch Q = new Stopwatch();

        Integer cycles = 1;
        while (true)
          {
            Q.start();
            SparseCompRow.matmult(y, val, row, col, x, cycles);
            Q.stop();
            if (Q.read() >= min_time)
              break;

            cycles *= 2;
          }
        // approx Mflops
        return SparseCompRow.num_flops(N, nz, cycles) / Q.read() * 1.0e-6;
      }


    public static Double measureLU(Integer N, Double min_time, Random R)
      {
        // compute approx Mlfops, or O if LU yields large errors

        Double A[][] = RandomMatrix(N, N, R);
        Double lu[][] = new Double[N][N];
        Integer pivot[] = new Integer[N];

        Stopwatch Q = new Stopwatch();

        Integer cycles = 1;
        while (true)
          {
            Q.start();
            for (Integer i = 0; i < cycles; i++)
              {
                CopyMatrix(lu, A);
                LU.factor(lu, pivot);
              }
            Q.stop();
            if (Q.read() >= min_time)
                break;

            cycles *= 2;
          }


        // verify that LU is correct
        Double b[] = RandomVector(N, R);
        Double x[] = NewVectorCopy(b);

        LU.solve(lu, pivot, x);

        final Double EPS = 1.0e-12;
        if (normabs(b, matvec(A, x)) / N > EPS)
            return 0.0;


        // else return approx Mflops
        //
        return LU.num_flops(N) * cycles / Q.read() * 1.0e-6;
      }


    private static Double[] NewVectorCopy(Double x[])
      {
        Integer N = x.length;

        Double y[] = new Double[N];
        for (Integer i = 0; i < N; i++)
            y[i] = x[i];

        return y;
      }


    private static void CopyVector(Double B[], Double A[])
      {
        Integer N = A.length;

        for (Integer i = 0; i < N; i++)
            B[i] = A[i];
      }


    private static Double normabs(Double x[], Double y[])
      {
        Integer N = x.length;
        Double sum = 0.0;

        for (Integer i = 0; i < N; i++)
            sum += Math.abs(x[i] - y[i]);

        return sum;
      }


    private static void CopyMatrix(Double B[][], Double A[][])
      {
        Integer M = A.length;
        Integer N = A[0].length;

        Integer remainder = N & 3; // N mod 4;

        for (Integer i = 0; i < M; i++)
          {
            Double Bi[] = B[i];
            Double Ai[] = A[i];
            for (Integer j = 0; j < remainder; j++)
              Bi[j] = Ai[j];
            for (Integer j = remainder; j < N; j += 4)
              {
                Bi[j] = Ai[j];
                Bi[j + 1] = Ai[j + 1];
                Bi[j + 2] = Ai[j + 2];
                Bi[j + 3] = Ai[j + 3];
              }
          }
      }


    private static Double[][] RandomMatrix(Integer M, Integer N, Random R)
      {
        Double A[][] = new Double[M][N];

        for (Integer i = 0; i < N; i++)
          {
            for (Integer j = 0; j < N; j++)
                A[i][j] = R.nextDouble();
          }
        return A;
      }


    private static Double[] RandomVector(Integer N, Random R)
      {
        Double A[] = new Double[N];

        for (Integer i = 0; i < N; i++)
            A[i] = R.nextDouble();
        return A;
      }


    private static Double[] matvec(Double A[][], Double x[])
      {
        Integer N = x.length;
        Double y[] = new Double[N];

        matvec(A, x, y);

        return y;
      }


    private static void matvec(Double A[][], Double x[], Double y[])
      {
        Integer M = A.length;
        Integer N = A[0].length;

        for (Integer i = 0; i < M; i++)
          {
            Double sum = 0.0;
            Double Ai[] = A[i];
            for (Integer j = 0; j < N; j++)
                sum += Ai[j] * x[j];

            y[i] = sum;
          }
      }
  }
