package com.softmoore.util;


/**
 * A simple "stopwatch" timer class with millisecond accuracy.<br>
 * Usage Notes:<br>
 * 1.  A stopwatch is started automatically when constructed.<br>
 * 2.  Method start() restarts the stopwatch.<br>
 * 3.  Method stop() can only be called if the stopwatch is currently running
 *     (i.e., if the stopwatch is not already stopped).<br>
 * 4.  A single stopwatch can be restarted/stopped multiple times.<br>
 * 5.  Methods that return the elapsed time (including the toString() method)
 *     can only be called if the stopwatch is stopped.<br>
 */
public class StopWatch
  {

    private long startTime;
    private long stopTime;

    private boolean isRunning;


    /**
     * Creates and starts the StopWatch.  Subsequent calls
     * to start simply reset the start time.
     */
    public StopWatch()
      {
        startTime = System.currentTimeMillis();
        stopTime  = 0;

        isRunning = true;
      }


    /**
     * Sets the start time for the StopWatch to the current time.
     */
    public void start()
      {
        startTime = System.currentTimeMillis();
        isRunning = true;
      }


    /**
     * Sets the stop time for the StopWatch to the current time.
     */
    public void stop()
      {
        if (!isRunning)
          {
            String message = "com.softmoore.util.Timer:  "
                           + "method stop() can only be called if the timer is currently running";
            throw new IllegalStateException(message);
          }

        stopTime = System.currentTimeMillis();

        isRunning = false;
      }


    /**
     * Returns the elapsed time in milliseconds
     */
    public long getElapsedTimeInMilliseconds()
      {
        if (isRunning)
          {
            String message = "com.softmoore.util.Timer:  "
                           + "can only be called if the timer has been stopped";
            throw new IllegalStateException(message);
          }

        return stopTime - startTime;
      }


    /**
     * Returns the elapsed time in seconds
     */
    public double getElapsedTimeInSeconds()
      {
        return ((double) getElapsedTimeInMilliseconds())/1000.0;
      }


    /**
     * Returns the elapsed time in minutes
     */
    public double getElapsedTimeInMinutes()
      {
        return getElapsedTimeInSeconds()/60.0;
      }


    /**
     * Returns the elapsed time in hours
     */
    public double getElapsedTimeInHours()
      {
        return getElapsedTimeInMinutes()/60.0;
      }


    public String toString()
      {
        return getElapsedTimeInMilliseconds() + " ms";
      }


    /**
     * Returns the start time in milliseconds
     */
    public long getStartTime()
      {
        return startTime;
      }


    /**
     * Returns the stop time in milliseconds
     */
    public long getStopTime()
      {
        return stopTime;
      }
  }
